# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Cosmos module wrapping the public and private key cryptography and ledger api."""

import base64
import hashlib
import json
import logging
import os
import subprocess  # nosec
import tempfile
import time
from collections import namedtuple
from pathlib import Path
from typing import Any, BinaryIO, Dict, List, Optional, Tuple

import requests
from bech32 import bech32_decode, bech32_encode, convertbits
from ecdsa import SECP256k1, SigningKey, VerifyingKey
from ecdsa.util import sigencode_string_canonize

from aea.common import Address
from aea.crypto.base import Crypto, FaucetApi, Helper, LedgerApi
from aea.exceptions import AEAEnforceError
from aea.helpers.base import try_decorator


_default_logger = logging.getLogger(__name__)

_COSMOS = "cosmos"
TESTNET_NAME = "testnet"
DEFAULT_FAUCET_URL = "INVALID_URL"
DEFAULT_ADDRESS = "INVALID_URL"
DEFAULT_CURRENCY_DENOM = "INVALID_CURRENCY_DENOM"
DEFAULT_CHAIN_ID = "INVALID_CHAIN_ID"


class CosmosHelper(Helper):
    """Helper class usable as Mixin for CosmosApi or as standalone class."""

    address_prefix = _COSMOS

    @staticmethod
    def is_transaction_settled(tx_receipt: Any) -> bool:
        """
        Check whether a transaction is settled or not.

        :param tx_digest: the digest associated to the transaction.
        :return: True if the transaction has been settled, False o/w.
        """
        is_successful = False
        if tx_receipt is not None:
            is_successful = True
        return is_successful

    @staticmethod
    def is_transaction_valid(
        tx: Any, seller: Address, client: Address, tx_nonce: str, amount: int,
    ) -> bool:
        """
        Check whether a transaction is valid or not.

        :param tx: the transaction.
        :param seller: the address of the seller.
        :param client: the address of the client.
        :param tx_nonce: the transaction nonce.
        :param amount: the amount we expect to get from the transaction.
        :return: True if the random_message is equals to tx['input']
        """
        if tx is None:
            return False  # pragma: no cover

        try:
            _tx = tx.get("tx").get("value").get("msg")[0]
            recovered_amount = int(_tx.get("value").get("amount")[0].get("amount"))
            sender = _tx.get("value").get("from_address")
            recipient = _tx.get("value").get("to_address")
            is_valid = (
                recovered_amount == amount and sender == client and recipient == seller
            )
        except (KeyError, IndexError):  # pragma: no cover
            is_valid = False
        return is_valid

    @staticmethod
    def generate_tx_nonce(seller: Address, client: Address) -> str:
        """
        Generate a unique hash to distinguish txs with the same terms.

        :param seller: the address of the seller.
        :param client: the address of the client.
        :return: return the hash in hex.
        """
        time_stamp = int(time.time())
        aggregate_hash = hashlib.sha256(
            b"".join([seller.encode(), client.encode(), time_stamp.to_bytes(32, "big")])
        )
        return aggregate_hash.hexdigest()

    @classmethod
    def get_address_from_public_key(cls, public_key: str) -> str:
        """
        Get the address from the public key.

        :param public_key: the public key
        :return: str
        """
        public_key_bytes = bytes.fromhex(public_key)
        s = hashlib.new("sha256", public_key_bytes).digest()
        r = hashlib.new("ripemd160", s).digest()
        five_bit_r = convertbits(r, 8, 5)
        if five_bit_r is None:  # pragma: nocover
            raise AEAEnforceError("Unsuccessful bech32.convertbits call")
        address = bech32_encode(cls.address_prefix, five_bit_r)
        return address

    @classmethod
    def recover_message(
        cls, message: bytes, signature: str, is_deprecated_mode: bool = False
    ) -> Tuple[Address, ...]:
        """
        Recover the addresses from the hash.

        :param message: the message we expect
        :param signature: the transaction signature
        :param is_deprecated_mode: if the deprecated signing was used
        :return: the recovered addresses
        """
        signature_b64 = base64.b64decode(signature)
        verifying_keys = VerifyingKey.from_public_key_recovery(
            signature_b64, message, SECP256k1, hashfunc=hashlib.sha256,
        )
        public_keys = [
            verifying_key.to_string("compressed").hex()
            for verifying_key in verifying_keys
        ]
        addresses = [
            cls.get_address_from_public_key(public_key) for public_key in public_keys
        ]
        return tuple(addresses)

    @staticmethod
    def get_hash(message: bytes) -> str:
        """
        Get the hash of a message.

        :param message: the message to be hashed.
        :return: the hash of the message.
        """
        digest = hashlib.sha256(message).hexdigest()
        return digest

    @classmethod
    def is_valid_address(cls, address: Address) -> bool:
        """
        Check if the address is valid.

        :param address: the address to validate
        """
        result = bech32_decode(address)
        return result != (None, None) and result[0] == cls.address_prefix


class CosmosCrypto(Crypto[SigningKey]):
    """Class wrapping the Account Generation from Ethereum ledger."""

    identifier = _COSMOS
    helper = CosmosHelper

    def __init__(self, private_key_path: Optional[str] = None):
        """
        Instantiate an ethereum crypto object.

        :param private_key_path: the private key path of the agent
        """
        super().__init__(private_key_path=private_key_path)
        self._public_key = self.entity.get_verifying_key().to_string("compressed").hex()
        self._address = self.helper.get_address_from_public_key(self.public_key)

    @property
    def private_key(self) -> str:
        """
        Return a private key.

        :return: a private key string
        """
        return self.entity.to_string().hex()

    @property
    def public_key(self) -> str:
        """
        Return a public key in hex format.

        :return: a public key string in hex format
        """
        return self._public_key

    @property
    def address(self) -> str:
        """
        Return the address for the key pair.

        :return: a display_address str
        """
        return self._address

    @classmethod
    def load_private_key_from_path(cls, file_name) -> SigningKey:
        """
        Load a private key in hex format from a file.

        :param file_name: the path to the hex file.
        :return: the Entity.
        """
        path = Path(file_name)
        with open(path, "r") as key:
            data = key.read()
            signing_key = SigningKey.from_string(bytes.fromhex(data), curve=SECP256k1)
        return signing_key

    def sign_message(  # pylint: disable=unused-argument
        self, message: bytes, is_deprecated_mode: bool = False
    ) -> str:
        """
        Sign a message in bytes string form.

        :param message: the message to be signed
        :param is_deprecated_mode: if the deprecated signing is used
        :return: signature of the message in string form
        """
        signature_compact = self.entity.sign_deterministic(
            message, hashfunc=hashlib.sha256, sigencode=sigencode_string_canonize,
        )
        signature_base64_str = base64.b64encode(signature_compact).decode("utf-8")
        return signature_base64_str

    @staticmethod
    def format_default_transaction(
        transaction: Any, signature: str, base64_pbk: str
    ) -> Any:
        """
        Format default CosmosSDK transaction and add signature.

        :param transaction: the transaction to be formatted
        :param signature: the transaction signature
        :param base64_pbk: the base64 formatted public key

        :return: formatted transaction with signature
        """
        pushable_tx = {
            "tx": {
                "msg": transaction["msgs"],
                "fee": transaction["fee"],
                "memo": transaction["memo"],
                "signatures": [
                    {
                        "signature": signature,
                        "pub_key": {
                            "type": "tendermint/PubKeySecp256k1",
                            "value": base64_pbk,
                        },
                        "account_number": transaction["account_number"],
                        "sequence": transaction["sequence"],
                    }
                ],
            },
            "mode": "async",
        }
        return pushable_tx

    @staticmethod
    def format_wasm_transaction(
        transaction: Any, signature: str, base64_pbk: str
    ) -> Any:
        """
        Format CosmWasm transaction and add signature.

        :param transaction: the transaction to be formatted
        :param signature: the transaction signature
        :param base64_pbk: the base64 formatted public key

        :return: formatted transaction with signature
        """
        pushable_tx = {
            "type": "cosmos-sdk/StdTx",
            "value": {
                "msg": transaction["msgs"],
                "fee": transaction["fee"],
                "signatures": [
                    {
                        "pub_key": {
                            "type": "tendermint/PubKeySecp256k1",
                            "value": base64_pbk,
                        },
                        "signature": signature,
                    }
                ],
                "memo": transaction["memo"],
            },
        }
        return pushable_tx

    def sign_transaction(self, transaction: Any) -> Any:
        """
        Sign a transaction in bytes string form.

        :param transaction: the transaction to be signed
        :return: signed transaction
        """
        transaction_str = json.dumps(transaction, separators=(",", ":"), sort_keys=True)
        transaction_bytes = transaction_str.encode("utf-8")
        signed_transaction = self.sign_message(transaction_bytes)
        base64_pbk = base64.b64encode(bytes.fromhex(self.public_key)).decode("utf-8")

        if (
            "msgs" in transaction
            and len(transaction["msgs"]) == 1
            and "type" in transaction["msgs"][0]
            and "wasm" in transaction["msgs"][0]["type"]
        ):
            return self.format_wasm_transaction(
                transaction, signed_transaction, base64_pbk
            )
        return self.format_default_transaction(
            transaction, signed_transaction, base64_pbk
        )

    @classmethod
    def generate_private_key(cls) -> SigningKey:
        """Generate a key pair for cosmos network."""
        signing_key = SigningKey.generate(curve=SECP256k1)
        return signing_key

    def dump(self, fp: BinaryIO) -> None:
        """
        Serialize crypto object as binary stream to `fp` (a `.write()`-supporting file-like object).

        :param fp: the output file pointer. Must be set in binary mode (mode='wb')
        :return: None
        """
        fp.write(self.private_key.encode("utf-8"))


class _CosmosApi(LedgerApi):
    """Class to interact with the Cosmos SDK via a HTTP APIs."""

    identifier = _COSMOS

    def __init__(self, **kwargs):
        """Initialize the Cosmos ledger APIs."""
        self._api = None
        self.network_address = kwargs.pop("address", DEFAULT_ADDRESS)
        self.denom = kwargs.pop("denom", DEFAULT_CURRENCY_DENOM)
        self.chain_id = kwargs.pop("chain_id", DEFAULT_CHAIN_ID)

    @property
    def api(self) -> None:
        """Get the underlying API object."""
        return self._api

    def get_balance(self, address: Address) -> Optional[int]:
        """Get the balance of a given account."""
        balance = self._try_get_balance(address)
        return balance

    @try_decorator(
        "Encountered exception when trying get balance: {}",
        logger_method=_default_logger.warning,
    )
    def _try_get_balance(self, address: Address) -> Optional[int]:
        """Try get the balance of a given account."""
        balance = None  # type: Optional[int]
        url = self.network_address + f"/bank/balances/{address}"
        response = requests.get(url=url)
        if response.status_code == 200:
            result = response.json()["result"]
            if len(result) == 0:
                balance = 0
            else:
                balance = int(result[0]["amount"])
        return balance

    def get_deploy_transaction(  # pylint: disable=arguments-differ
        self,
        contract_interface: Dict[str, str],
        deployer_address: Address,
        tx_fee: int = 0,
        gas: int = 80000,
        denom: Optional[str] = None,
        memo: str = "",
        chain_id: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Create a CosmWasm bytecode deployment transaction.

        :param sender_address: the sender address of the message initiator.
        :param filename: the path to wasm bytecode file.
        :param gas: Maximum amount of gas to be used on executing command.
        :param memo: Any string comment.
        :param chain_id: the Chain ID of the CosmWasm transaction. Default is 1 (i.e. mainnet).
        :return: the unsigned CosmWasm contract deploy message
        """
        denom = denom if denom is not None else self.denom
        chain_id = chain_id if chain_id is not None else self.chain_id
        account_number, sequence = self._try_get_account_number_and_sequence(
            deployer_address
        )
        deploy_msg = {
            "type": "wasm/store-code",
            "value": {
                "sender": deployer_address,
                "wasm_byte_code": contract_interface["wasm_byte_code"],
                "source": "",
                "builder": "",
            },
        }
        tx = self._get_transaction(
            account_number,
            chain_id,
            tx_fee,
            denom,
            gas,
            memo,
            sequence,
            msg=deploy_msg,
        )
        return tx

    def get_init_transaction(
        self,
        deployer_address: Address,
        code_id: int,
        init_msg: Any,
        amount: int,
        tx_fee: int,
        gas: int = 80000,
        denom: Optional[str] = None,
        label: str = "",
        memo: str = "",
        chain_id: Optional[str] = None,
    ) -> Optional[Any]:
        """
        Create a CosmWasm InitMsg transaction.

        :param deployer_address: the deployer address of the message initiator.
        :param amount: Contract's initial funds amount
        :param code_id: the ID of contract bytecode.
        :param init_msg: the InitMsg containing parameters for contract constructor.
        :param gas: Maximum amount of gas to be used on executing command.
        :param denom: the name of the denomination of the contract funds
        :param label: the label name of the contract
        :param memo: Any string comment.
        :param chain_id: the Chain ID of the CosmWasm transaction. Default is 1 (i.e. mainnet).
        :return: the unsigned CosmWasm InitMsg
        """
        denom = denom if denom is not None else self.denom
        chain_id = chain_id if chain_id is not None else self.chain_id
        account_number, sequence = self._try_get_account_number_and_sequence(
            deployer_address
        )
        instantiate_msg = {
            "type": "wasm/instantiate",
            "value": {
                "sender": deployer_address,
                "code_id": str(code_id),
                "label": label,
                "init_msg": init_msg,
                "init_funds": [{"denom": denom, "amount": str(amount)}],
            },
        }
        tx = self._get_transaction(
            account_number,
            chain_id,
            tx_fee,
            denom,
            gas,
            memo,
            sequence,
            msg=instantiate_msg,
        )
        return tx

    def get_handle_transaction(
        self,
        sender_address: Address,
        contract_address: Address,
        handle_msg: Any,
        amount: int,
        tx_fee: int,
        denom: Optional[str] = None,
        gas: int = 80000,
        memo: str = "",
        chain_id: Optional[str] = None,
    ) -> Optional[Any]:
        """
        Create a CosmWasm HandleMsg transaction.

        :param sender_address: the sender address of the message initiator.
        :param contract_address: the address of the smart contract.
        :param handle_msg: HandleMsg in JSON format.
        :param gas: Maximum amount of gas to be used on executing command.
        :param memo: Any string comment.
        :param chain_id: the Chain ID of the CosmWasm transaction. Default is 1 (i.e. mainnet).
        :return: the unsigned CosmWasm HandleMsg
        """
        denom = denom if denom is not None else self.denom
        chain_id = chain_id if chain_id is not None else self.chain_id
        account_number, sequence = self._try_get_account_number_and_sequence(
            sender_address
        )
        execute_msg = {
            "type": "wasm/execute",
            "value": {
                "sender": sender_address,
                "contract": contract_address,
                "msg": handle_msg,
                "sent_funds": [{"amount": str(amount), "denom": denom}],
            },
        }
        tx = self._get_transaction(
            account_number,
            chain_id,
            tx_fee,
            denom,
            gas,
            memo,
            sequence,
            msg=execute_msg,
        )
        return tx

    @staticmethod
    @try_decorator(
        "Encountered exception when trying to execute wasm transaction: {}",
        logger_method=_default_logger.warning,
    )
    def try_execute_wasm_transaction(
        tx_signed: Any, signed_tx_filename: str = "tx.signed"
    ) -> Optional[str]:
        """
        Execute a CosmWasm Transaction. QueryMsg doesn't require signing.

        :param tx_signed: the signed transaction.
        :return: the transaction digest
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(os.path.join(tmpdirname, signed_tx_filename), "w") as f:
                f.write(json.dumps(tx_signed))

            command = [
                "wasmcli",
                "tx",
                "broadcast",
                os.path.join(tmpdirname, signed_tx_filename),
            ]

            stdout, _ = subprocess.Popen(  # nosec
                command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            ).communicate()

        return stdout.decode("ascii")

    @staticmethod
    @try_decorator(
        "Encountered exception when trying to execute wasm query: {}",
        logger_method=_default_logger.warning,
    )
    def try_execute_wasm_query(
        contract_address: Address, query_msg: Any
    ) -> Optional[str]:
        """
        Execute a CosmWasm QueryMsg. QueryMsg doesn't require signing.

        :param contract_address: the address of the smart contract.
        :param query_msg: QueryMsg in JSON format.
        :return: the message receipt
        """
        command = [
            "wasmcli",
            "query",
            "wasm",
            "contract-state",
            "smart",
            str(contract_address),
            json.dumps(query_msg),
        ]

        stdout, _ = subprocess.Popen(  # nosec
            command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        ).communicate()

        return stdout.decode("ascii")

    def get_transfer_transaction(  # pylint: disable=arguments-differ
        self,
        sender_address: Address,
        destination_address: Address,
        amount: int,
        tx_fee: int,
        tx_nonce: str,
        denom: Optional[str] = None,
        gas: int = 80000,
        memo: str = "",
        chain_id: Optional[str] = None,
        **kwargs,
    ) -> Optional[Any]:
        """
        Submit a transfer transaction to the ledger.

        :param sender_address: the sender address of the payer.
        :param destination_address: the destination address of the payee.
        :param amount: the amount of wealth to be transferred.
        :param tx_fee: the transaction fee.
        :param tx_nonce: verifies the authenticity of the tx
        :param denom: the denomination of tx fee and amount
        :param gas: the gas used.
        :param memo: memo to include in tx.
        :param chain_id: the chain ID of the transaction.
        :return: the transfer transaction
        """
        denom = denom if denom is not None else self.denom
        chain_id = chain_id if chain_id is not None else self.chain_id
        account_number, sequence = self._try_get_account_number_and_sequence(
            sender_address
        )
        transfer_msg = {
            "type": "cosmos-sdk/MsgSend",
            "value": {
                "amount": [{"amount": str(amount), "denom": denom}],
                "from_address": sender_address,
                "to_address": destination_address,
            },
        }
        tx = self._get_transaction(
            account_number,
            chain_id,
            tx_fee,
            denom,
            gas,
            memo,
            sequence,
            msg=transfer_msg,
        )
        return tx

    @staticmethod
    def _get_transaction(
        account_number: int,
        chain_id: str,
        tx_fee: int,
        denom: str,
        gas: int,
        memo: str,
        sequence: int,
        msg: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Get a transaction.

        :param account_number: the account number.
        :param chain_id: the chain ID of the transaction.
        :param tx_fee: the transaction fee.
        :param denom: the denomination of tx fee and amount
        :param gas: the gas used.
        :param memo: memo to include in tx.
        :param msg: the transaction msg.
        :param sequence: the sequence.
        :return: the transaction
        """
        tx = {
            "account_number": str(account_number),
            "chain_id": chain_id,
            "fee": {
                "amount": [{"amount": str(tx_fee), "denom": denom}],
                "gas": str(gas),
            },
            "memo": memo,
            "msgs": [msg],
            "sequence": str(sequence),
        }
        return tx

    @try_decorator(
        "Encountered exception when trying to get account number and sequence: {}",
        logger_method=_default_logger.warning,
    )
    def _try_get_account_number_and_sequence(
        self, address: Address
    ) -> Optional[Tuple[int, int]]:
        """
        Try get account number and sequence for an address.

        :param address: the address
        :return: a tuple of account number and sequence
        """
        result = None  # type: Optional[Tuple[int, int]]
        url = self.network_address + f"/auth/accounts/{address}"
        response = requests.get(url=url)
        if response.status_code == 200:
            result = (
                int(response.json()["result"]["value"]["account_number"]),
                int(response.json()["result"]["value"]["sequence"]),
            )
        return result

    def send_signed_transaction(self, tx_signed: Any) -> Optional[str]:
        """
        Send a signed transaction and wait for confirmation.

        :param tx_signed: the signed transaction
        :return: tx_digest, if present
        """
        if self.is_cosmwasm_transaction(tx_signed):
            tx_digest = self.try_execute_wasm_transaction(tx_signed)
        elif self.is_transfer_transaction(tx_signed):
            tx_digest = self._try_send_signed_transaction(tx_signed)
        else:  # pragma: nocover
            _default_logger.warning(
                "Cannot send transaction. Unknown transaction type: {}".format(
                    tx_signed
                )
            )
            tx_digest = None
        return tx_digest

    @staticmethod
    def is_cosmwasm_transaction(tx_signed: Any) -> bool:
        """Check whether it is a cosmwasm tx."""
        try:
            _type = tx_signed["value"]["msg"][0]["type"]
            result = _type in ["wasm/store-code", "wasm/instantiate", "wasm/execute"]
        except KeyError:  # pragma: nocover
            result = False
        return result

    @staticmethod
    def is_transfer_transaction(tx_signed: Any) -> bool:
        """Check whether it is a transfer tx."""
        try:
            _type = tx_signed["tx"]["msg"][0]["type"]
            result = _type in ["cosmos-sdk/MsgSend"]
        except KeyError:  # pragma: nocover
            result = False
        return result

    @try_decorator(
        "Encountered exception when trying to send tx: {}",
        logger_method=_default_logger.warning,
    )
    def _try_send_signed_transaction(self, tx_signed: Any) -> Optional[str]:
        """
        Try send the signed transaction.

        :param tx_signed: the signed transaction
        :return: tx_digest, if present
        """
        tx_digest = None  # type: Optional[str]
        url = self.network_address + "/txs"
        response = requests.post(url=url, json=tx_signed)
        if response.status_code == 200:
            tx_digest = response.json()["txhash"]
        else:  # pragma: nocover
            _default_logger.error("Cannot send transaction: {}".format(response.json()))
        return tx_digest

    def get_transaction_receipt(self, tx_digest: str) -> Optional[Any]:
        """
        Get the transaction receipt for a transaction digest.

        :param tx_digest: the digest associated to the transaction.
        :return: the tx receipt, if present
        """
        tx_receipt = self._try_get_transaction_receipt(tx_digest)
        return tx_receipt

    @try_decorator(
        "Encountered exception when trying to get transaction receipt: {}",
        logger_method=_default_logger.warning,
    )
    def _try_get_transaction_receipt(self, tx_digest: str) -> Optional[Any]:
        """
        Try get the transaction receipt for a transaction digest.

        :param tx_digest: the digest associated to the transaction.
        :return: the tx receipt, if present
        """
        result = None  # type: Optional[Any]
        url = self.network_address + f"/txs/{tx_digest}"
        response = requests.get(url=url)
        if response.status_code == 200:
            result = response.json()
        return result

    def get_transaction(self, tx_digest: str) -> Optional[Any]:
        """
        Get the transaction for a transaction digest.

        :param tx_digest: the digest associated to the transaction.
        :return: the tx, if present
        """
        # Cosmos does not distinguis between transaction receipt and transaction
        tx_receipt = self._try_get_transaction_receipt(tx_digest)
        return tx_receipt

    def get_contract_instance(
        self, contract_interface: Dict[str, str], contract_address: Optional[str] = None
    ) -> Any:
        """
        Get the instance of a contract.

        :param contract_interface: the contract interface.
        :param contract_address: the contract address.
        :return: the contract instance
        """
        # Instance object not available for cosmwasm
        return None

    @staticmethod
    def _execute_shell_command(command: List[str]) -> List[Dict[str, str]]:
        """
        Execute command using subprocess and get result as JSON dict.

        :param command: the shell command to be executed
        :return: the stdout result converted to JSON dict
        """
        stdout, _ = subprocess.Popen(  # nosec
            command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        ).communicate()

        return json.loads(stdout.decode("ascii"))

    def get_last_code_id(self) -> int:
        """
        Get ID of latest deployed .wasm bytecode.

        :return: code id of last deployed .wasm bytecode
        """

        command = ["wasmcli", "query", "wasm", "list-code"]
        res = self._execute_shell_command(command)

        return int(res[-1]["id"])

    def get_contract_address(self, code_id: int) -> str:
        """
        Get contract address of latest initialised contract by its ID.

        :param code_id: id of deployed CosmWasm bytecode
        :return: contract address of last initialised contract
        """

        command = ["wasmcli", "query", "wasm", "list-contract-by-code", str(code_id)]
        res = self._execute_shell_command(command)

        return res[-1]["address"]


class CosmosApi(_CosmosApi, CosmosHelper):
    """Class to interact with the Cosmos SDK via a HTTP APIs."""


""" Equivalent to:

@dataclass
class CosmosFaucetStatus:
    tx_digest: Optional[str]
    status: str
    status_code: int
"""
CosmosFaucetStatus = namedtuple(
    "CosmosFaucetStatus", ["tx_digest", "status", "status_code"]
)


class CosmosFaucetApi(FaucetApi):
    """Cosmos testnet faucet API."""

    FAUCET_STATUS_PENDING = 1  # noqa: F841
    FAUCET_STATUS_PROCESSING = 2  # noqa: F841
    FAUCET_STATUS_COMPLETED = 20  # noqa: F841
    FAUCET_STATUS_FAILED = 21  # noqa: F841
    FAUCET_STATUS_TIMED_OUT = 22  # noqa: F841
    FAUCET_STATUS_RATE_LIMITED = 23  # noqa: F841
    FAUCET_STATUS_RATE_UNAVAILABLE = 99  # noqa: F841

    identifier = _COSMOS
    testnet_faucet_url = DEFAULT_FAUCET_URL
    testnet_name = TESTNET_NAME

    def __init__(self, poll_interval=None):
        """Initialize CosmosFaucetApi."""
        self._poll_interval = float(poll_interval or 1)

    def get_wealth(self, address: Address) -> None:
        """
        Get wealth from the faucet for the provided address.

        :param address: the address.
        :return: None
        :raises: RuntimeError of explicit faucet failures
        """
        uid = self._try_create_faucet_claim(address)
        if uid is None:  # pragma: nocover
            raise RuntimeError("Unable to create faucet claim")

        while True:

            # lookup status form the claim uid
            status = self._try_check_faucet_claim(uid)
            if status is None:  # pragma: nocover
                raise RuntimeError("Failed to check faucet claim status")

            # if the status is complete
            if status.status_code == self.FAUCET_STATUS_COMPLETED:
                break

            # if the status is failure
            if status.status_code > self.FAUCET_STATUS_COMPLETED:  # pragma: nocover
                raise RuntimeError(f"Failed to get wealth for {address}")

            # if the status is incomplete
            time.sleep(self._poll_interval)

    @classmethod
    @try_decorator(
        "An error occured while attempting to request a faucet request:\n{}",
        logger_method=_default_logger.error,
    )
    def _try_create_faucet_claim(cls, address: Address) -> Optional[str]:
        """
        Create a token faucet claim request

        :param address: the address to request funds
        :return: None on failure, otherwise the request uid
        """
        response = requests.post(
            url=cls._faucet_request_uri(), data={"Address": address}
        )

        uid = None
        if response.status_code == 200:
            data = response.json()
            uid = data["uid"]

            _default_logger.info("Wealth claim generated, uid: {}".format(uid))
        else:  # pragma: no cover
            _default_logger.warning(
                "Response: {}, Text: {}".format(response.status_code, response.text)
            )

        return uid

    @classmethod
    @try_decorator(
        "An error occured while attempting to request a faucet request:\n{}",
        logger_method=_default_logger.error,
    )
    def _try_check_faucet_claim(cls, uid: str) -> Optional[CosmosFaucetStatus]:
        """
        Check the status of a faucet request

        :param uid: The request uid to be checked
        :return: None on failure otherwise a CosmosFaucetStatus for the specified uid
        """
        response = requests.get(cls._faucet_status_uri(uid))
        if response.status_code != 200:  # pragma: nocover
            _default_logger.warning(
                "Response: {}, Text: {}".format(response.status_code, response.text)
            )
            return None

        # parse the response
        data = response.json()
        return CosmosFaucetStatus(
            tx_digest=data.get("txDigest"),
            status=data["status"],
            status_code=data["statusCode"],
        )

    @classmethod
    def _faucet_request_uri(cls) -> str:
        """Generates the request URI derived from `cls.faucet_base_url`."""
        if cls.testnet_faucet_url is None:  # pragma: nocover
            raise ValueError("Testnet faucet url not set.")
        return f"{cls.testnet_faucet_url}/claim/requests"

    @classmethod
    def _faucet_status_uri(cls, uid: str) -> str:
        """Generates the status URI derived from `cls.faucet_base_url`."""
        return f"{cls._faucet_request_uri()}/{uid}"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the behaviours."""
from collections import Counter
from unittest import TestCase, mock

import pytest

from aea.skills.behaviours import (
    CyclicBehaviour,
    FSMBehaviour,
    OneShotBehaviour,
    SequenceBehaviour,
    State,
    TickerBehaviour,
)


def test_sequence_behaviour():
    """Test the sequence behaviour."""
    outputs = []

    class MySequenceBehaviour(SequenceBehaviour):
        """Custom sequence behaviour."""

        def setup(self) -> None:
            """Setup behaviour."""
            pass

        def teardown(self) -> None:
            """Teardown behaviour."""
            pass

    class SimpleOneShotBehaviour(OneShotBehaviour):
        """Custom simple behaviour."""

        def __init__(self, name, **kwargs):
            super().__init__(name=name, **kwargs)

        def setup(self) -> None:
            """Setup behaviour."""
            pass

        def teardown(self) -> None:
            """Teardown behaviour."""
            pass

        def act(self) -> None:
            """Act implementation."""
            outputs.append(self.name)

    # TODO let the initialization of a behaviour action from constructor
    a = SimpleOneShotBehaviour("a", skill_context=object())
    b = SimpleOneShotBehaviour("b", skill_context=object())
    c = SimpleOneShotBehaviour("c", skill_context=object())
    sequence = MySequenceBehaviour([a, b, c], name="abc", skill_context=object())

    max_iterations = 10
    i = 0
    while not sequence.is_done() and i < max_iterations:
        sequence.act()
        i += 1

    assert outputs == ["a", "b", "c"]


def test_act_parameter():
    """Test the 'act' parameter."""
    counter = Counter(i=0)

    def increment_counter(counter=counter):
        counter += Counter(i=1)

    assert counter["i"] == 0

    one_shot_behaviour = OneShotBehaviour(
        act=lambda: increment_counter(), skill_context=object(), name="my_behaviour"
    )
    one_shot_behaviour.act()
    assert counter["i"] == 1


class SimpleFSMBehaviour(FSMBehaviour):
    """A Finite-State Machine behaviour for testing purposes."""

    def setup(self) -> None:
        """Setup behaviour."""
        pass

    def teardown(self) -> None:
        """Teardown behaviour."""
        pass


class SimpleStateBehaviour(State):
    """A simple state behaviour to be added in a FSMBehaviour."""

    def __init__(self, shared_list, event_to_trigger=None, **kwargs):
        """Initialise simple behaviour."""
        super().__init__(**kwargs)
        self.shared_list = shared_list
        self.event_to_trigger = event_to_trigger
        self.executed = False

    def setup(self) -> None:
        """Setup behaviour."""
        pass

    def teardown(self) -> None:
        """Teardown behaviour."""
        pass

    def act(self) -> None:
        """Act implementation."""
        self.shared_list.append(self.name)
        self.executed = True
        self._event = self.event_to_trigger

    def is_done(self) -> bool:
        """Get is done."""
        return self.executed


def test_fms_behaviour():
    """Test the finite-state machine behaviour."""
    outputs = []

    a = SimpleStateBehaviour(
        outputs, name="a", event_to_trigger="move_to_b", skill_context=object()
    )
    b = SimpleStateBehaviour(
        outputs, name="b", event_to_trigger="move_to_c", skill_context=object()
    )
    c = SimpleStateBehaviour(outputs, name="c", skill_context=object())
    fsm = SimpleFSMBehaviour(name="abc", skill_context=object())
    fsm.register_state(str(a.name), a, initial=True)
    fsm.register_state(str(b.name), b)
    fsm.register_final_state(str(c.name), c)
    fsm.register_transition("a", "b", "move_to_b")
    fsm.register_transition("b", "c", "move_to_c")

    max_iterations = 10
    i = 0
    while not fsm.is_done() and i < max_iterations:
        fsm.act()
        i += 1

    assert outputs == ["a", "b", "c"]


class TestFSMBehaviourCreation:
    """Test FSMBehaviour creation."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.fsm_behaviour = SimpleFSMBehaviour(name="fsm", skill_context=object())
        cls.outputs = []
        cls.a = SimpleStateBehaviour(cls.outputs, name="a", skill_context=object())
        cls.b = SimpleStateBehaviour(cls.outputs, name="b", skill_context=object())
        cls.c = SimpleStateBehaviour(cls.outputs, name="c", skill_context=object())

    def test_initial_state_is_none(self):
        """Test that the initial state is None."""
        assert self.fsm_behaviour.initial_state is None

    def test_states_is_empty(self):
        """Test that the states is an empty set."""
        assert self.fsm_behaviour.states == set()

    def test_final_states_is_empty(self):
        """Test that the final states is an empty set."""
        assert self.fsm_behaviour.final_states == set()

    def test_add_and_remove_state(self):
        """Test that adding and removing a state works correctly."""
        assert self.fsm_behaviour.states == set()
        self.fsm_behaviour.register_state("a", self.a)
        assert self.fsm_behaviour.states == {"a"}
        self.fsm_behaviour.unregister_state("a")
        assert self.fsm_behaviour.states == set()

    def test_add_and_remove_initial_state(self):
        """Test that adding and removing an initial state works correctly."""
        assert self.fsm_behaviour.states == set()
        self.fsm_behaviour.register_state("a", self.a, initial=True)
        assert self.fsm_behaviour.states == {"a"}
        assert self.fsm_behaviour.initial_state == "a"
        assert self.fsm_behaviour.final_states == set()
        self.fsm_behaviour.unregister_state("a")
        assert self.fsm_behaviour.states == set()
        assert self.fsm_behaviour.initial_state is None
        assert self.fsm_behaviour.final_states == set()

    def test_add_and_remove_final_state(self):
        """Test that adding and removing final states works correctly."""
        assert self.fsm_behaviour.states == set()
        self.fsm_behaviour.register_final_state("a", self.a)
        assert self.fsm_behaviour.states == {"a"}
        assert self.fsm_behaviour.initial_state is None
        assert self.fsm_behaviour.final_states == {"a"}
        self.fsm_behaviour.unregister_state("a")
        assert self.fsm_behaviour.states == set()
        assert self.fsm_behaviour.initial_state is None
        assert self.fsm_behaviour.final_states == set()

    def test_register_initial_state_twice(self):
        """Test that the register state with initial=True works correctly when called twice."""
        assert self.fsm_behaviour.initial_state is None
        self.fsm_behaviour.register_state("a", self.a, initial=True)
        assert self.fsm_behaviour.initial_state == "a"
        self.fsm_behaviour.register_state("b", self.b, initial=True)
        assert self.fsm_behaviour.initial_state == "b"
        self.fsm_behaviour.unregister_state("a")
        self.fsm_behaviour.unregister_state("b")
        assert self.fsm_behaviour.initial_state is None

    def test_register_twice_same_state(self):
        """Test that registering twice a state with the same name raises an error."""
        self.fsm_behaviour.register_state("a", self.a)
        with pytest.raises(ValueError, match="State name already existing."):
            self.fsm_behaviour.register_state("a", self.a)
        self.fsm_behaviour.unregister_state("a")

    def test_register_transition(self):
        """Test register transition."""
        self.fsm_behaviour.register_transition("state_1", "state_2")
        self.fsm_behaviour.register_transition("state_1", "state_2", "an_event")
        assert self.fsm_behaviour.transitions == {
            "state_1": {None: "state_2", "an_event": "state_2"}
        }
        self.fsm_behaviour.unregister_transition("state_1", "state_2", None)
        self.fsm_behaviour.unregister_transition("state_1", "state_2", "an_event")
        assert self.fsm_behaviour.transitions == dict()

    def test_register_same_transition_twice(self):
        """Test that when we try to register twice the same transition we raise an error."""
        self.fsm_behaviour.register_transition("state_1", "state_2")
        with pytest.raises(ValueError, match="Transition already registered."):
            self.fsm_behaviour.register_transition("state_1", "state_2")
        self.fsm_behaviour.unregister_transition("state_1", "state_2")

        self.fsm_behaviour.register_transition("state_1", "state_2", "an_event")
        with pytest.raises(ValueError, match="Transition already registered."):
            self.fsm_behaviour.register_transition("state_1", "state_2", "an_event")
        self.fsm_behaviour.unregister_transition("state_1", "state_2", "an_event")


class CyclicBehaviourTestCase(TestCase):
    """Test case for CyclicBehaviour class."""

    def setUp(self):
        """Set the test up."""

        class TestCyclicBehaviour(CyclicBehaviour):
            """Class for testing CyclicBehaviour abstract class."""

            def setup(self, *args):
                """Set up."""
                pass

            def teardown(self, *args):
                """Tear down."""
                pass

        self.TestCyclicBehaviour = TestCyclicBehaviour

    def test_init_positive(self):
        """Test for init positive result."""
        self.TestCyclicBehaviour(skill_context="skill_context", name="name")

    def test_act_wrapper_positive(self):
        """Test for act_wrapper positive result."""
        obj = self.TestCyclicBehaviour(skill_context="skill_context", name="name")
        obj.act = mock.Mock()
        assert obj.number_of_executions == 0
        obj.act_wrapper()
        obj.act.assert_called_once()
        assert obj.number_of_executions == 1


class TickerBehaviourTestCase(TestCase):
    """Test case for TickerBehaviour class."""

    def setUp(self):
        """Set the test up."""

        class TestTickerBehaviour(TickerBehaviour):
            """Class for testing TickerBehaviour abstract class."""

            def setup(self, *args):
                """Set up."""
                pass

            def teardown(self, *args):
                """Tear down."""
                pass

        self.TestTickerBehaviour = TestTickerBehaviour

    def test_init_positive(self):
        """Test for init positive result."""
        self.TestTickerBehaviour(skill_context="skill_context", name="name")

    def test_tick_interval_positive(self):
        """Test for tick_interval property positive result."""
        obj = self.TestTickerBehaviour(skill_context="skill_context", name="name")
        obj.tick_interval

    def test_start_at_positive(self):
        """Test for start_at property positive result."""
        obj = self.TestTickerBehaviour(skill_context="skill_context", name="name")
        obj.start_at

    def test_last_act_time_positive(self):
        """Test for last_act_time property positive result."""
        obj = self.TestTickerBehaviour(skill_context="skill_context", name="name")
        obj.last_act_time

    def test_act_wrapper_positive(self):
        """Test for act_wrapper positive result."""
        obj = self.TestTickerBehaviour(skill_context="skill_context", name="name")
        obj.is_done = mock.Mock(return_value=False)
        obj.is_time_to_act = mock.Mock(return_value=True)
        obj.act = mock.Mock()
        obj.act_wrapper()
        obj.act.assert_called_once()

    def test_is_time_to_act_positive(self):
        """Test for is_time_to_act positive result."""
        obj = self.TestTickerBehaviour(skill_context="skill_context", name="name")
        obj.is_time_to_act()


class FSMBehaviourTestCase(TestCase):
    """Test case for FSMBehaviour class."""

    def setUp(self):
        """Set the test up."""

        class TestFSMBehaviour(FSMBehaviour):
            """Class for testing FSMBehaviour abstract class."""

            def setup(self, *args):
                """Set up."""
                pass

            def teardown(self, *args):
                """Tear down."""
                pass

        self.TestFSMBehaviour = TestFSMBehaviour

    def test_is_started_positive(self):
        """Test for is_started property positive result."""
        obj = self.TestFSMBehaviour(skill_context="skill_context", name="name")
        obj.is_started

    def test_register_final_state_already_exists(self):
        """Test for register_final_state already exists."""
        obj = self.TestFSMBehaviour(skill_context="skill_context", name="name")
        obj._name_to_state = ["name"]
        with self.assertRaises(ValueError):
            obj.register_final_state("name", "state")

    def test_unregister_state_not_exists(self):
        """Test for unregister_state not exists."""
        obj = self.TestFSMBehaviour(skill_context="skill_context", name="name")
        obj._name_to_state = []
        with self.assertRaises(ValueError):
            obj.unregister_state("name")

    def test_initial_state_not_state(self):
        """Test for initial_state not a state."""
        obj = self.TestFSMBehaviour(skill_context="skill_context", name="name")
        obj._name_to_state = []
        with self.assertRaises(ValueError):
            obj.initial_state = "name"

    def test_initial_state_positive(self):
        """Test for initial_state positive result."""
        obj = self.TestFSMBehaviour(skill_context="skill_context", name="name")
        obj._name_to_state = ["name"]
        obj.initial_state = "name"

    def test_act_no_current(self):
        """Test for act method no current state."""
        obj = self.TestFSMBehaviour(skill_context="skill_context", name="name")
        obj.current = None
        obj.act()

    def test_act_no_current_got(self):
        """Test for act method no current state got."""
        obj = self.TestFSMBehaviour(skill_context="skill_context", name="name")
        obj.get_state = mock.Mock(return_value=None)
        obj.current = "current"
        obj.act()
        obj.get_state.assert_called_once()

    def test_act_current_in_final_states(self):
        """Test for act method current in final_states."""
        obj = self.TestFSMBehaviour(skill_context="skill_context", name="name")
        current_state = mock.Mock()
        current_state.act_wrapper = mock.Mock()
        current_state.is_done = mock.Mock(return_value=True)

        obj.final_states.add(current_state)
        obj.get_state = mock.Mock(return_value=current_state)
        obj.current = "current"
        obj.act()

        obj.get_state.assert_called_once()
        current_state.act_wrapper.assert_called_once()
        current_state.is_done.assert_called_once()

    def test_unregister_transition_value_error(self):
        """Test for unregister_transition method ValueError raises."""
        obj = self.TestFSMBehaviour(skill_context="skill_context", name="name")
        obj.transitions = {}
        with self.assertRaises(ValueError):
            obj.unregister_transition("source", "destination")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------


"""HTTP server connection, channel, server, and handler."""
import asyncio
import email
import logging
from abc import ABC, abstractmethod
from asyncio import CancelledError
from asyncio.events import AbstractEventLoop
from asyncio.futures import Future
from traceback import format_exc
from typing import Dict, Optional, Set, cast
from urllib.parse import parse_qs, urlencode, urlparse

from aiohttp import web
from aiohttp.web_request import BaseRequest
from openapi_core import create_spec
from openapi_core.validation.request.datatypes import OpenAPIRequest, RequestParameters
from openapi_core.validation.request.shortcuts import validate_request
from openapi_core.validation.request.validators import RequestValidator
from openapi_spec_validator.exceptions import (  # pylint: disable=wrong-import-order
    OpenAPIValidationError,
)
from openapi_spec_validator.schemas import (  # pylint: disable=wrong-import-order
    read_yaml_file,
)
from werkzeug.datastructures import (  # pylint: disable=wrong-import-order
    ImmutableMultiDict,
)

from aea.common import Address
from aea.configurations.base import PublicId
from aea.connections.base import Connection, ConnectionStates
from aea.mail.base import Envelope, EnvelopeContext, Message, URI
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import DialogueLabel

from packages.fetchai.protocols.http.dialogues import HttpDialogue
from packages.fetchai.protocols.http.dialogues import HttpDialogues as BaseHttpDialogues
from packages.fetchai.protocols.http.message import HttpMessage


SUCCESS = 200
NOT_FOUND = 404
REQUEST_TIMEOUT = 408
SERVER_ERROR = 500

_default_logger = logging.getLogger("aea.packages.fetchai.connections.http_server")

RequestId = DialogueLabel
PUBLIC_ID = PublicId.from_str("fetchai/http_server:0.10.0")


class HttpDialogues(BaseHttpDialogues):
    """The dialogues class keeps track of all http dialogues."""

    def __init__(self, self_address: Address, **kwargs) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            # The server connection maintains the dialogue on behalf of the client
            return HttpDialogue.Role.CLIENT

        BaseHttpDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            **kwargs,
        )


def headers_to_string(headers: Dict):
    """
    Convert headers to string.

    :param headers: dict

    :return: str
    """
    msg = email.message.Message()
    for name, value in headers.items():
        msg.add_header(name, value)
    return msg.as_string()


class Request(OpenAPIRequest):
    """Generic request object."""

    @property
    def is_id_set(self):
        """Check if id is set."""
        return self._id is not None

    @property
    def id(self) -> RequestId:
        """Get the request id."""
        return self._id

    @id.setter
    def id(self, request_id: RequestId) -> None:
        """Set the request id."""
        self._id = request_id

    @classmethod
    async def create(cls, http_request: BaseRequest) -> "Request":
        """
        Create a request.

        :param http_request: http_request
        :return: a request
        """
        method = http_request.method.lower()

        parsed_path = urlparse(http_request.path_qs)

        url = http_request.url

        body = await http_request.read()

        mimetype = http_request.content_type

        query_params = parse_qs(parsed_path.query, keep_blank_values=True)

        parameters = RequestParameters(
            query=ImmutableMultiDict(query_params),
            header=headers_to_string(dict(http_request.headers)),
            path={},
        )

        request = Request(
            full_url_pattern=str(url),
            method=method,
            parameters=parameters,
            body=body,
            mimetype=mimetype,
        )
        return request

    def to_envelope_and_set_id(
        self, connection_id: PublicId, agent_address: str, dialogues: HttpDialogues,
    ) -> Envelope:
        """
        Process incoming API request by packaging into Envelope and sending it in-queue.

        :param connection_id: id of the connection
        :param agent_address: agent's address
        :param dialogue_reference: new dialog refernece for envelope

        :return: envelope
        """
        url = (
            self.full_url_pattern
            if self.parameters.query == {}
            else self.full_url_pattern + "?" + urlencode(self.parameters.query)
        )
        uri = URI(self.full_url_pattern)
        context = EnvelopeContext(connection_id=connection_id, uri=uri)
        http_message, http_dialogue = dialogues.create(
            counterparty=agent_address,
            performative=HttpMessage.Performative.REQUEST,
            method=self.method,
            url=url,
            headers=self.parameters.header,
            bodyy=self.body if self.body is not None else b"",
            version="",
        )
        dialogue = cast(HttpDialogue, http_dialogue)
        self.id = dialogue.incomplete_dialogue_label
        envelope = Envelope(
            to=http_message.to,
            sender=http_message.sender,
            protocol_id=http_message.protocol_id,
            context=context,
            message=http_message,
        )
        return envelope


class Response(web.Response):
    """Generic response object."""

    @classmethod
    def from_message(cls, http_message: HttpMessage) -> "Response":
        """
        Turn an envelope into a response.

        :param http_message: the http_message
        :return: the response
        """
        if http_message.performative == HttpMessage.Performative.RESPONSE:

            if http_message.is_set("headers") and http_message.headers:
                headers: Optional[dict] = dict(
                    email.message_from_string(http_message.headers).items()
                )
            else:
                headers = None

            response = cls(
                status=http_message.status_code,
                reason=http_message.status_text,
                body=http_message.bodyy,
                headers=headers,
            )
        else:  # pragma: nocover
            response = cls(status=SERVER_ERROR, text="Server error")
        return response


class APISpec:
    """API Spec class to verify a request against an OpenAPI/Swagger spec."""

    def __init__(
        self,
        api_spec_path: Optional[str] = None,
        server: Optional[str] = None,
        logger: logging.Logger = _default_logger,
    ):
        """
        Initialize the API spec.

        :param api_spec_path: Directory API path and filename of the API spec YAML source file.
        """
        self._validator = None  # type: Optional[RequestValidator]
        self.logger = logger
        if api_spec_path is not None:
            try:
                api_spec_dict = read_yaml_file(api_spec_path)
                if server is not None:
                    api_spec_dict["servers"] = [{"url": server}]
                api_spec = create_spec(api_spec_dict)
                self._validator = RequestValidator(api_spec)
            except OpenAPIValidationError as e:  # pragma: nocover
                self.logger.error(
                    f"API specification YAML source file not correctly formatted: {str(e)}"
                )
            except Exception:
                self.logger.exception(
                    "API specification YAML source file not correctly formatted."
                )
                raise

    def verify(self, request: Request) -> bool:
        """
        Verify a http_method, url and param against the provided API spec.

        :param request: the request object
        :return: whether or not the request conforms with the API spec
        """
        if self._validator is None:
            self.logger.debug("Skipping API verification!")
            return True

        try:
            validate_request(self._validator, request)
        except Exception:  # pragma: nocover # pylint: disable=broad-except
            self.logger.exception("APISpec verify error")
            return False
        return True


class BaseAsyncChannel(ABC):
    """Base asynchronous channel class."""

    def __init__(self, address: Address, connection_id: PublicId,) -> None:
        """
        Initialize a channel.

        :param address: the address of the agent.
        :param connection_id: public id of connection using this chanel.
        """
        self._in_queue = None  # type: Optional[asyncio.Queue]
        self._loop = None  # type: Optional[asyncio.AbstractEventLoop]
        self.is_stopped = True
        self.address = address
        self.connection_id = connection_id

    @abstractmethod
    async def connect(self, loop: AbstractEventLoop) -> None:
        """
        Connect.

        Upon HTTP Channel connection, kickstart the HTTP Server in its own thread.

        :param loop: asyncio event loop

        :return: None
        """
        self._loop = loop
        self._in_queue = asyncio.Queue()
        self.is_stopped = False

    async def get_message(self) -> Optional["Envelope"]:
        """
        Get http response from in-queue.

        :return: None or envelope with http response.
        """
        if self._in_queue is None:
            raise ValueError("Looks like channel is not connected!")

        try:
            return await self._in_queue.get()
        except CancelledError:  # pragma: nocover
            return None

    @abstractmethod
    def send(self, envelope: Envelope) -> None:
        """
        Send the envelope in_queue.

        :param envelope: the envelope
        :return: None
        """

    @abstractmethod
    async def disconnect(self) -> None:
        """
        Disconnect.

        Shut-off the HTTP Server.
        """


class HTTPChannel(BaseAsyncChannel):
    """A wrapper for an RESTful API with an internal HTTPServer."""

    RESPONSE_TIMEOUT = 300

    def __init__(
        self,
        address: Address,
        host: str,
        port: int,
        api_spec_path: Optional[str],
        connection_id: PublicId,
        restricted_to_protocols: Set[PublicId],
        timeout_window: float = 5.0,
        logger: logging.Logger = _default_logger,
    ):
        """
        Initialize a channel and process the initial API specification from the file path (if given).

        :param address: the address of the agent.
        :param host: RESTful API hostname / IP address
        :param port: RESTful API port number
        :param api_spec_path: Directory API path and filename of the API spec YAML source file.
        :param connection_id: public id of connection using this chanel.
        :param restricted_to_protocols: set of restricted protocols
        :param timeout_window: the timeout (in seconds) for a request to be handled.
        """
        super().__init__(address=address, connection_id=connection_id)
        self.host = host
        self.port = port
        self.server_address = "http://{}:{}".format(self.host, self.port)
        self.restricted_to_protocols = restricted_to_protocols

        self._api_spec = APISpec(api_spec_path, self.server_address, logger)
        self.timeout_window = timeout_window
        self.http_server: Optional[web.TCPSite] = None
        self.pending_requests: Dict[RequestId, Future] = {}
        self._dialogues = HttpDialogues(str(HTTPServerConnection.connection_id))
        self.logger = logger

    @property
    def api_spec(self) -> APISpec:
        """Get the api spec."""
        return self._api_spec

    async def connect(self, loop: AbstractEventLoop) -> None:
        """
        Connect.

        Upon HTTP Channel connection, kickstart the HTTP Server in its own thread.

        :param loop: asyncio event loop

        :return: None
        """
        if self.is_stopped:
            await super().connect(loop)

            try:
                await self._start_http_server()
                self.logger.info(
                    "HTTP Server has connected to port: {}.".format(self.port)
                )
            except Exception:  # pragma: nocover # pylint: disable=broad-except
                self.is_stopped = True
                self._in_queue = None
                self.logger.exception(
                    "Failed to start server on {}:{}.".format(self.host, self.port)
                )

    async def _http_handler(self, http_request: BaseRequest) -> Response:
        """
        Verify the request then send the request to Agent as an envelope.

        :param request: the request object

        :return: a tuple of response code and response description
        """
        request = await Request.create(http_request)
        if self._in_queue is None:  # pragma: nocover
            raise ValueError("Channel not connected!")

        is_valid_request = self.api_spec.verify(request)

        if not is_valid_request:
            self.logger.warning(f"request is not valid: {request}")
            return Response(status=NOT_FOUND, reason="Request Not Found")

        try:
            # turn request into envelope
            envelope = request.to_envelope_and_set_id(
                self.connection_id, self.address, dialogues=self._dialogues,
            )

            self.pending_requests[request.id] = Future()

            # send the envelope to the agent's inbox (via self.in_queue)
            await self._in_queue.put(envelope)
            # wait for response envelope within given timeout window (self.timeout_window) to appear in dispatch_ready_envelopes

            response_message = await asyncio.wait_for(
                self.pending_requests[request.id], timeout=self.RESPONSE_TIMEOUT,
            )

            return Response.from_message(response_message)

        except asyncio.TimeoutError:
            return Response(status=REQUEST_TIMEOUT, reason="Request Timeout")
        except BaseException:  # pragma: nocover # pylint: disable=broad-except
            self.logger.exception("Error during handling incoming request")
            return Response(
                status=SERVER_ERROR, reason="Server Error", text=format_exc()
            )
        finally:
            if request.is_id_set:
                self.pending_requests.pop(request.id, None)

    async def _start_http_server(self) -> None:
        """Start http server."""
        server = web.Server(self._http_handler)
        runner = web.ServerRunner(server)
        await runner.setup()
        self.http_server = web.TCPSite(runner, self.host, self.port)
        await self.http_server.start()

    def send(self, envelope: Envelope) -> None:
        """
        Send the envelope in_queue.

        :param envelope: the envelope
        :return: None
        """
        if self.http_server is None:  # pragma: nocover
            raise ValueError("Server not connected, call connect first!")

        if envelope.protocol_id not in self.restricted_to_protocols:
            self.logger.error(
                "This envelope cannot be sent with the http connection: protocol_id={}".format(
                    envelope.protocol_id
                )
            )
            raise ValueError("Cannot send message.")

        message = cast(HttpMessage, envelope.message)
        dialogue = self._dialogues.update(message)

        if dialogue is None:
            self.logger.warning(
                "Could not create dialogue for message={}".format(message)
            )
            return

        future = self.pending_requests.pop(dialogue.incomplete_dialogue_label, None)

        if not future:
            self.logger.warning(
                "Dropping message={} for incomplete_dialogue_label={} which has timed out.".format(
                    message, dialogue.incomplete_dialogue_label
                )
            )
        else:
            future.set_result(message)

    async def disconnect(self) -> None:
        """
        Disconnect.

        Shut-off the HTTP Server.
        """
        if self.http_server is None:  # pragma: nocover
            raise ValueError("Server not connected, call connect first!")

        if not self.is_stopped:
            await self.http_server.stop()
            self.logger.info("HTTP Server has shutdown on port: {}.".format(self.port))
            self.is_stopped = True
            self._in_queue = None


class HTTPServerConnection(Connection):
    """Proxy to the functionality of the http server implementing a RESTful API specification."""

    connection_id = PUBLIC_ID

    def __init__(self, **kwargs):
        """Initialize a HTTP server connection."""
        super().__init__(**kwargs)
        host = cast(str, self.configuration.config.get("host"))
        port = cast(int, self.configuration.config.get("port"))
        if host is None or port is None:  # pragma: nocover
            raise ValueError("host and port must be set!")
        api_spec_path = cast(str, self.configuration.config.get("api_spec_path"))
        self.channel = HTTPChannel(
            self.address,
            host,
            port,
            api_spec_path,
            connection_id=self.connection_id,
            restricted_to_protocols=self.restricted_to_protocols,
            logger=self.logger,
        )

    async def connect(self) -> None:
        """
        Connect to the http.

        :return: None
        """
        if self.is_connected:
            return

        self._state.set(ConnectionStates.connecting)
        self.channel.logger = self.logger
        await self.channel.connect(loop=self.loop)
        if self.channel.is_stopped:
            self._state.set(ConnectionStates.disconnected)
        else:
            self._state.set(ConnectionStates.connected)

    async def disconnect(self) -> None:
        """
        Disconnect from HTTP.

        :return: None
        """
        if self.is_disconnected:
            return

        self._state.set(ConnectionStates.disconnecting)
        await self.channel.disconnect()
        self._state.set(ConnectionStates.disconnected)

    async def send(self, envelope: "Envelope") -> None:
        """
        Send an envelope.

        :param envelope: the envelop
        :return: None
        """
        self._ensure_connected()
        self.channel.send(envelope)

    async def receive(self, *args, **kwargs) -> Optional["Envelope"]:
        """
        Receive an envelope.

        :return: the envelope received, or None.
        """
        self._ensure_connected()
        try:
            return await self.channel.get_message()
        except CancelledError:  # pragma: no cover
            return None
